#![no_std]
#![feature(asm_experimental_arch)]

use core::{arch::asm, marker::PhantomData};

use arduino_hal::port::{mode::Output, Pin, PinOps};
use avr_device::interrupt::free;

use smart_leds::{SmartLedsWrite, RGB8};
use static_pins::StaticPinOps;

const MSB: u8 = 0x80;

#[repr(transparent)]
pub struct Neopixel<P> {
    _pin: PhantomData<Pin<Output, P>>,
}

impl<P> Neopixel<P>
where
    P: PinOps + StaticPinOps,
{
    #[inline]
    pub fn new(_pin: Pin<Output, P>) -> Self {
        Self {
            _pin: PhantomData {},
        }
    }

    fn write_color_channel(&self, data: u8) {
        let mut data = data;
        unsafe {
            for _ in 0..8 {
                P::set_high();
                asm!("rjmp +0");
                if data & MSB == 0 {
                    P::set_low();
                }
                asm!(
                    "
                    rjmp +0
                    rjmp +0
                    "
                );
                P::set_low();
                asm!(
                    "
                    rjmp +0
                    rjmp +0
                    rjmp +0
                    "
                );
                data <<= 1;
            }
        }
    }
}

impl<P> SmartLedsWrite for Neopixel<P>
where
    P: PinOps + StaticPinOps,
{
    type Error = ();
    type Color = smart_leds::RGB<u8>;
    fn write<T, I>(&mut self, iterator: T) -> Result<(), Self::Error>
    where
        T: IntoIterator<Item = I>,
        I: Into<Self::Color>,
    {
        free(|_cs| {
            for i in iterator {
                let RGB8 { r, g, b } = i.into();
                for value in [g, r, b] {
                    self.write_color_channel(value);
                }
            }
        });
        Ok(())
    }
}
